package de.tudresden.inf.st.jastadd.testDumper;

import de.tudresden.inf.st.jastadd.dumpAst.ast.DumpNode;
import org.jastadd.testDumper.ast.C;
import org.jastadd.testDumper.ast.D;
import org.jastadd.testDumper.ast.Root;
import org.jastadd.testDumper.ast.SubC;
import org.junit.jupiter.api.Test;

import java.util.List;
import java.util.Optional;

import static de.tudresden.inf.st.jastadd.testDumper.TestUtils.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

public class TestSimple {

  @Test
  public void testEmpty() {
    Root root = createRoot(null, null);

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertEquals(1, actualRoot.getNumDumpToken());
    assertEquals(0, actualRoot.getNumDumpChildNode());
    assertEquals(0, actualRoot.getNumDumpRelation());
  }

  @Test void testChildlessNonterminal() {
    Root root = createRoot(createA(A_NAME, a -> a.setD(new D())), null);
    List<DumpNode> nodes = TestUtils.dumpModel(root);
    Optional<DumpNode> optionalCDumpNode = nodes.stream().filter(n -> n.getObject() instanceof D).findFirst();
    assertTrue(optionalCDumpNode.isPresent());
    assertTrue(optionalCDumpNode.get().isAstNode());
  }

  @Test
  public void testOneNormalChild() {
    Root root = createRoot(createA(A_NAME), null);

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertEquals(1, actualRoot.getNumDumpToken());
    assertEquals(1, actualRoot.getNumDumpChildNode());
    assertEquals(0, actualRoot.getNumDumpRelation());
    assertThatMapOf(normalChildren(actualRoot)).containsExactlyInAnyOrder(tuple("A", A_NAME));
  }

  @Test
  public void testListChildren() {
    Root root = createRoot(null, null, createB(B1_NAME), createB(B2_NAME), createB(B3_NAME));

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, B1_NAME, B2_NAME, B3_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertEquals(1, actualRoot.getNumDumpToken());
    assertEquals(1, actualRoot.getNumDumpChildNode());
    assertEquals(0, actualRoot.getNumDumpRelation());
    assertThatMapOf(listChildren(actualRoot), "B").containsExactlyInAnyOrder(B1_NAME, B2_NAME, B3_NAME);
  }

  @Test
  public void testOneOptChild() {
    Root root = createRoot(null, createC(C_NAME));

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, C_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertEquals(1, actualRoot.getNumDumpToken());
    assertEquals(1, actualRoot.getNumDumpChildNode());
    assertEquals(0, actualRoot.getNumDumpRelation());
    assertThatMapOf(normalChildren(actualRoot)).containsExactlyInAnyOrder(tuple("C", C_NAME));
  }

  @Test
  public void testNormalUniRelation() {
    Root root = createRoot(createA(A_NAME), createC(C_NAME), createB(B_NAME));
    root.getB(0).setOneA(root.getA());

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, B_NAME, C_NAME);
    DumpNode actualB = TestUtils.findByName(nodes, B_NAME);
    assertThatMapOf(normalRelationChildren(actualB)).containsExactlyInAnyOrder(tuple("OneA", A_NAME));
  }

  @Test
  public void testListUniRelation() {
    final String A2_NAME = "anotherA";
    Root root = createRoot(createA(A_NAME), createC(C_NAME), createB(B_NAME));
    root.getC().setA(createA(A2_NAME));
    root.getB(0).addManyA(root.getA());
    root.getB(0).addManyA(root.getC().getA());

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, A2_NAME, B_NAME, C_NAME);
    DumpNode actualB = TestUtils.findByName(nodes, B_NAME);
    assertThatMapOf(listRelationChildren(actualB), "ManyA").containsExactlyInAnyOrder(A_NAME, A2_NAME);
  }

  @Test
  public void testOptUniRelation() {
    Root root = createRoot(createA(A_NAME), createC(C_NAME), createB(B_NAME));
    root.getB(0).setMaybeC(root.getC());

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, B_NAME, C_NAME);
    DumpNode actualB = TestUtils.findByName(nodes, B_NAME);
    assertThatMapOf(normalRelationChildren(actualB)).containsExactlyInAnyOrder(tuple("MaybeC", C_NAME));
  }

  @Test
  public void testNormalBiRelation() {
    Root root = createRoot(createA(A_NAME), createC(C_NAME), createB(B_NAME));
    root.getC().setBiA1(root.getA());

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, B_NAME, C_NAME);
    // bidirectional relations are currently not found, instead there will be two unidirectional ones
    DumpNode actualA = TestUtils.findByName(nodes, A_NAME);
    assertThatMapOf(normalRelationChildren(actualA)).containsExactlyInAnyOrder(tuple("BiC1", C_NAME));
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThatMapOf(normalRelationChildren(actualC)).containsExactlyInAnyOrder(tuple("BiA1", A_NAME));
  }

  @Test
  public void testListBiRelation() {
    Root root = createRoot(createA(A_NAME), createC(C_NAME), createB(B_NAME));
    root.getC().addBiA2(root.getA());

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, B_NAME, C_NAME);
    // bidirectional relations are currently not found, instead there will be two unidirectional ones
    DumpNode actualA = TestUtils.findByName(nodes, A_NAME);
    assertThatMapOf(normalRelationChildren(actualA)).containsExactlyInAnyOrder(tuple("BiC2", C_NAME));
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThatMapOf(listRelationChildren(actualC), "BiA2").containsExactlyInAnyOrder(A_NAME);
  }

  @Test
  public void testOptBiRelation() {
    Root root = createRoot(createA(A_NAME), createC(C_NAME), createB(B_NAME));
    root.getC().setBiA3(root.getA());

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, B_NAME, C_NAME);
    // bidirectional relations are currently not found, instead there will be two unidirectional ones
    DumpNode actualA = TestUtils.findByName(nodes, A_NAME);
    assertThatMapOf(normalRelationChildren(actualA)).containsExactlyInAnyOrder(tuple("BiC3", C_NAME));
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThatMapOf(normalRelationChildren(actualC)).containsExactlyInAnyOrder(tuple("BiA3", A_NAME));
  }

  @Test
  public void testOneNormalReferenceToken() {
    Root root = createRoot(createA(A_NAME), createC(C_NAME), createB(B_NAME));
    root.getC().setRawReference(root.getA());

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, B_NAME, C_NAME);
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThatMapOf(referenceTokens(actualC)).containsExactlyInAnyOrder(tuple(TOKEN_LABEL_RAW_REFERENCE, A_NAME));
  }

  @Test
  public void testChildrenReorder1() {
    C subC = new SubC();
    subC.setName(C_NAME);
    subC.setUnwanted(5);
    Root root = createRoot(createA(A_NAME), subC, createB(B_NAME));

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, B_NAME, C_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertEquals(1, actualRoot.getNumDumpToken());
    assertEquals(3, actualRoot.getNumDumpChildNode());
    assertEquals(0, actualRoot.getNumDumpRelation());
    assertThat(valueTokens(findByName(nodes, C_NAME))).containsOnly(
        entry("Name", C_NAME),
        entry(TOKEN_LABEL_UNWANTED, 5));
  }

}
